r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Taskrouter
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, List, Optional, Union
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class WorkspaceRealTimeStatisticsInstance(InstanceResource):
    """
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Workspace resource.
    :ivar activity_statistics: The number of current Workers by Activity.
    :ivar longest_task_waiting_age: The age of the longest waiting Task.
    :ivar longest_task_waiting_sid: The SID of the longest waiting Task.
    :ivar tasks_by_priority: The number of Tasks by priority. For example: `{\"0\": \"10\", \"99\": \"5\"}` shows 10 Tasks at priority 0 and 5 at priority 99.
    :ivar tasks_by_status: The number of Tasks by their current status. For example: `{\"pending\": \"1\", \"reserved\": \"3\", \"assigned\": \"2\", \"completed\": \"5\"}`.
    :ivar total_tasks: The total number of Tasks.
    :ivar total_workers: The total number of Workers in the Workspace.
    :ivar workspace_sid: The SID of the Workspace.
    :ivar url: The absolute URL of the Workspace statistics resource.
    """

    def __init__(self, version: Version, payload: Dict[str, Any], workspace_sid: str):
        super().__init__(version)

        self.account_sid: Optional[str] = payload.get("account_sid")
        self.activity_statistics: Optional[List[Dict[str, object]]] = payload.get(
            "activity_statistics"
        )
        self.longest_task_waiting_age: Optional[int] = deserialize.integer(
            payload.get("longest_task_waiting_age")
        )
        self.longest_task_waiting_sid: Optional[str] = payload.get(
            "longest_task_waiting_sid"
        )
        self.tasks_by_priority: Optional[Dict[str, object]] = payload.get(
            "tasks_by_priority"
        )
        self.tasks_by_status: Optional[Dict[str, object]] = payload.get(
            "tasks_by_status"
        )
        self.total_tasks: Optional[int] = deserialize.integer(
            payload.get("total_tasks")
        )
        self.total_workers: Optional[int] = deserialize.integer(
            payload.get("total_workers")
        )
        self.workspace_sid: Optional[str] = payload.get("workspace_sid")
        self.url: Optional[str] = payload.get("url")

        self._solution = {
            "workspace_sid": workspace_sid,
        }
        self._context: Optional[WorkspaceRealTimeStatisticsContext] = None

    @property
    def _proxy(self) -> "WorkspaceRealTimeStatisticsContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: WorkspaceRealTimeStatisticsContext for this WorkspaceRealTimeStatisticsInstance
        """
        if self._context is None:
            self._context = WorkspaceRealTimeStatisticsContext(
                self._version,
                workspace_sid=self._solution["workspace_sid"],
            )
        return self._context

    def fetch(
        self, task_channel: Union[str, object] = values.unset
    ) -> "WorkspaceRealTimeStatisticsInstance":
        """
        Fetch the WorkspaceRealTimeStatisticsInstance

        :param task_channel: Only calculate real-time statistics on this TaskChannel. Can be the TaskChannel's SID or its `unique_name`, such as `voice`, `sms`, or `default`.

        :returns: The fetched WorkspaceRealTimeStatisticsInstance
        """
        return self._proxy.fetch(
            task_channel=task_channel,
        )

    async def fetch_async(
        self, task_channel: Union[str, object] = values.unset
    ) -> "WorkspaceRealTimeStatisticsInstance":
        """
        Asynchronous coroutine to fetch the WorkspaceRealTimeStatisticsInstance

        :param task_channel: Only calculate real-time statistics on this TaskChannel. Can be the TaskChannel's SID or its `unique_name`, such as `voice`, `sms`, or `default`.

        :returns: The fetched WorkspaceRealTimeStatisticsInstance
        """
        return await self._proxy.fetch_async(
            task_channel=task_channel,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Taskrouter.V1.WorkspaceRealTimeStatisticsInstance {}>".format(
            context
        )


class WorkspaceRealTimeStatisticsContext(InstanceContext):

    def __init__(self, version: Version, workspace_sid: str):
        """
        Initialize the WorkspaceRealTimeStatisticsContext

        :param version: Version that contains the resource
        :param workspace_sid: The SID of the Workspace to fetch.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "workspace_sid": workspace_sid,
        }
        self._uri = "/Workspaces/{workspace_sid}/RealTimeStatistics".format(
            **self._solution
        )

    def fetch(
        self, task_channel: Union[str, object] = values.unset
    ) -> WorkspaceRealTimeStatisticsInstance:
        """
        Fetch the WorkspaceRealTimeStatisticsInstance

        :param task_channel: Only calculate real-time statistics on this TaskChannel. Can be the TaskChannel's SID or its `unique_name`, such as `voice`, `sms`, or `default`.

        :returns: The fetched WorkspaceRealTimeStatisticsInstance
        """

        data = values.of(
            {
                "TaskChannel": task_channel,
            }
        )

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(
            method="GET", uri=self._uri, params=data, headers=headers
        )

        return WorkspaceRealTimeStatisticsInstance(
            self._version,
            payload,
            workspace_sid=self._solution["workspace_sid"],
        )

    async def fetch_async(
        self, task_channel: Union[str, object] = values.unset
    ) -> WorkspaceRealTimeStatisticsInstance:
        """
        Asynchronous coroutine to fetch the WorkspaceRealTimeStatisticsInstance

        :param task_channel: Only calculate real-time statistics on this TaskChannel. Can be the TaskChannel's SID or its `unique_name`, such as `voice`, `sms`, or `default`.

        :returns: The fetched WorkspaceRealTimeStatisticsInstance
        """

        data = values.of(
            {
                "TaskChannel": task_channel,
            }
        )

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )

        return WorkspaceRealTimeStatisticsInstance(
            self._version,
            payload,
            workspace_sid=self._solution["workspace_sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Taskrouter.V1.WorkspaceRealTimeStatisticsContext {}>".format(
            context
        )


class WorkspaceRealTimeStatisticsList(ListResource):

    def __init__(self, version: Version, workspace_sid: str):
        """
        Initialize the WorkspaceRealTimeStatisticsList

        :param version: Version that contains the resource
        :param workspace_sid: The SID of the Workspace to fetch.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "workspace_sid": workspace_sid,
        }

    def get(self) -> WorkspaceRealTimeStatisticsContext:
        """
        Constructs a WorkspaceRealTimeStatisticsContext

        """
        return WorkspaceRealTimeStatisticsContext(
            self._version, workspace_sid=self._solution["workspace_sid"]
        )

    def __call__(self) -> WorkspaceRealTimeStatisticsContext:
        """
        Constructs a WorkspaceRealTimeStatisticsContext

        """
        return WorkspaceRealTimeStatisticsContext(
            self._version, workspace_sid=self._solution["workspace_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Taskrouter.V1.WorkspaceRealTimeStatisticsList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Api
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import date
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, serialize, values

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page
from twilio.rest.api.v2010.account.usage.record.all_time import AllTimeList
from twilio.rest.api.v2010.account.usage.record.daily import DailyList
from twilio.rest.api.v2010.account.usage.record.last_month import LastMonthList
from twilio.rest.api.v2010.account.usage.record.monthly import MonthlyList
from twilio.rest.api.v2010.account.usage.record.this_month import ThisMonthList
from twilio.rest.api.v2010.account.usage.record.today import TodayList
from twilio.rest.api.v2010.account.usage.record.yearly import YearlyList
from twilio.rest.api.v2010.account.usage.record.yesterday import YesterdayList


class RecordInstance(InstanceResource):
    """
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that accrued the usage.
    :ivar api_version: The API version used to create the resource.
    :ivar as_of: Usage records up to date as of this timestamp, formatted as YYYY-MM-DDTHH:MM:SS+00:00. All timestamps are in GMT
    :ivar category: The category of usage. For more information, see [Usage Categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
    :ivar count: The number of usage events, such as the number of calls.
    :ivar count_unit: The units in which `count` is measured, such as `calls` for calls or `messages` for SMS.
    :ivar description: A plain-language description of the usage category.
    :ivar end_date: The last date for which usage is included in the UsageRecord. The date is specified in GMT and formatted as `YYYY-MM-DD`.
    :ivar price: The total price of the usage in the currency specified in `price_unit` and associated with the account.
    :ivar price_unit: The currency in which `price` is measured, in [ISO 4127](https://www.iso.org/iso/home/standards/currency_codes.htm) format, such as `usd`, `eur`, and `jpy`.
    :ivar start_date: The first date for which usage is included in this UsageRecord. The date is specified in GMT and formatted as `YYYY-MM-DD`.
    :ivar subresource_uris: A list of related resources identified by their URIs. For more information, see [List Subresources](https://www.twilio.com/docs/usage/api/usage-record#list-subresources).
    :ivar uri: The URI of the resource, relative to `https://api.twilio.com`.
    :ivar usage: The amount used to bill usage and measured in units described in `usage_unit`.
    :ivar usage_unit: The units in which `usage` is measured, such as `minutes` for calls or `messages` for SMS.
    """

    def __init__(self, version: Version, payload: Dict[str, Any], account_sid: str):
        super().__init__(version)

        self.account_sid: Optional[str] = payload.get("account_sid")
        self.api_version: Optional[str] = payload.get("api_version")
        self.as_of: Optional[str] = payload.get("as_of")
        self.category: Optional[str] = payload.get("category")
        self.count: Optional[str] = payload.get("count")
        self.count_unit: Optional[str] = payload.get("count_unit")
        self.description: Optional[str] = payload.get("description")
        self.end_date: Optional[date] = deserialize.iso8601_date(
            payload.get("end_date")
        )
        self.price: Optional[float] = deserialize.decimal(payload.get("price"))
        self.price_unit: Optional[str] = payload.get("price_unit")
        self.start_date: Optional[date] = deserialize.iso8601_date(
            payload.get("start_date")
        )
        self.subresource_uris: Optional[Dict[str, object]] = payload.get(
            "subresource_uris"
        )
        self.uri: Optional[str] = payload.get("uri")
        self.usage: Optional[str] = payload.get("usage")
        self.usage_unit: Optional[str] = payload.get("usage_unit")

        self._solution = {
            "account_sid": account_sid,
        }

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.RecordInstance {}>".format(context)


class RecordPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> RecordInstance:
        """
        Build an instance of RecordInstance

        :param payload: Payload response from the API
        """
        return RecordInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.RecordPage>"


class RecordList(ListResource):

    def __init__(self, version: Version, account_sid: str):
        """
        Initialize the RecordList

        :param version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the UsageRecord resources to read.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
        }
        self._uri = "/Accounts/{account_sid}/Usage/Records.json".format(
            **self._solution
        )

        self._all_time: Optional[AllTimeList] = None
        self._daily: Optional[DailyList] = None
        self._last_month: Optional[LastMonthList] = None
        self._monthly: Optional[MonthlyList] = None
        self._this_month: Optional[ThisMonthList] = None
        self._today: Optional[TodayList] = None
        self._yearly: Optional[YearlyList] = None
        self._yesterday: Optional[YesterdayList] = None

    def stream(
        self,
        category: Union[str, object] = values.unset,
        start_date: Union[date, object] = values.unset,
        end_date: Union[date, object] = values.unset,
        include_subaccounts: Union[bool, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[RecordInstance]:
        """
        Streams RecordInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str category: The [usage category](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) of the UsageRecord resources to read. Only UsageRecord resources in the specified category are retrieved.
        :param date start_date: Only include usage that has occurred on or after this date. Specify the date in GMT and format as `YYYY-MM-DD`. You can also specify offsets from the current date, such as: `-30days`, which will set the start date to be 30 days before the current date.
        :param date end_date: Only include usage that occurred on or before this date. Specify the date in GMT and format as `YYYY-MM-DD`.  You can also specify offsets from the current date, such as: `+30days`, which will set the end date to 30 days from the current date.
        :param bool include_subaccounts: Whether to include usage from the master account and all its subaccounts. Can be: `true` (the default) to include usage from the master account and all subaccounts or `false` to retrieve usage from only the specified account.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(
            category=category,
            start_date=start_date,
            end_date=end_date,
            include_subaccounts=include_subaccounts,
            page_size=limits["page_size"],
        )

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        category: Union[str, object] = values.unset,
        start_date: Union[date, object] = values.unset,
        end_date: Union[date, object] = values.unset,
        include_subaccounts: Union[bool, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[RecordInstance]:
        """
        Asynchronously streams RecordInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str category: The [usage category](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) of the UsageRecord resources to read. Only UsageRecord resources in the specified category are retrieved.
        :param date start_date: Only include usage that has occurred on or after this date. Specify the date in GMT and format as `YYYY-MM-DD`. You can also specify offsets from the current date, such as: `-30days`, which will set the start date to be 30 days before the current date.
        :param date end_date: Only include usage that occurred on or before this date. Specify the date in GMT and format as `YYYY-MM-DD`.  You can also specify offsets from the current date, such as: `+30days`, which will set the end date to 30 days from the current date.
        :param bool include_subaccounts: Whether to include usage from the master account and all its subaccounts. Can be: `true` (the default) to include usage from the master account and all subaccounts or `false` to retrieve usage from only the specified account.
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            category=category,
            start_date=start_date,
            end_date=end_date,
            include_subaccounts=include_subaccounts,
            page_size=limits["page_size"],
        )

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        category: Union[str, object] = values.unset,
        start_date: Union[date, object] = values.unset,
        end_date: Union[date, object] = values.unset,
        include_subaccounts: Union[bool, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[RecordInstance]:
        """
        Lists RecordInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str category: The [usage category](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) of the UsageRecord resources to read. Only UsageRecord resources in the specified category are retrieved.
        :param date start_date: Only include usage that has occurred on or after this date. Specify the date in GMT and format as `YYYY-MM-DD`. You can also specify offsets from the current date, such as: `-30days`, which will set the start date to be 30 days before the current date.
        :param date end_date: Only include usage that occurred on or before this date. Specify the date in GMT and format as `YYYY-MM-DD`.  You can also specify offsets from the current date, such as: `+30days`, which will set the end date to 30 days from the current date.
        :param bool include_subaccounts: Whether to include usage from the master account and all its subaccounts. Can be: `true` (the default) to include usage from the master account and all subaccounts or `false` to retrieve usage from only the specified account.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                category=category,
                start_date=start_date,
                end_date=end_date,
                include_subaccounts=include_subaccounts,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        category: Union[str, object] = values.unset,
        start_date: Union[date, object] = values.unset,
        end_date: Union[date, object] = values.unset,
        include_subaccounts: Union[bool, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[RecordInstance]:
        """
        Asynchronously lists RecordInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str category: The [usage category](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) of the UsageRecord resources to read. Only UsageRecord resources in the specified category are retrieved.
        :param date start_date: Only include usage that has occurred on or after this date. Specify the date in GMT and format as `YYYY-MM-DD`. You can also specify offsets from the current date, such as: `-30days`, which will set the start date to be 30 days before the current date.
        :param date end_date: Only include usage that occurred on or before this date. Specify the date in GMT and format as `YYYY-MM-DD`.  You can also specify offsets from the current date, such as: `+30days`, which will set the end date to 30 days from the current date.
        :param bool include_subaccounts: Whether to include usage from the master account and all its subaccounts. Can be: `true` (the default) to include usage from the master account and all subaccounts or `false` to retrieve usage from only the specified account.
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                category=category,
                start_date=start_date,
                end_date=end_date,
                include_subaccounts=include_subaccounts,
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        category: Union[str, object] = values.unset,
        start_date: Union[date, object] = values.unset,
        end_date: Union[date, object] = values.unset,
        include_subaccounts: Union[bool, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> RecordPage:
        """
        Retrieve a single page of RecordInstance records from the API.
        Request is executed immediately

        :param category: The [usage category](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) of the UsageRecord resources to read. Only UsageRecord resources in the specified category are retrieved.
        :param start_date: Only include usage that has occurred on or after this date. Specify the date in GMT and format as `YYYY-MM-DD`. You can also specify offsets from the current date, such as: `-30days`, which will set the start date to be 30 days before the current date.
        :param end_date: Only include usage that occurred on or before this date. Specify the date in GMT and format as `YYYY-MM-DD`.  You can also specify offsets from the current date, such as: `+30days`, which will set the end date to 30 days from the current date.
        :param include_subaccounts: Whether to include usage from the master account and all its subaccounts. Can be: `true` (the default) to include usage from the master account and all subaccounts or `false` to retrieve usage from only the specified account.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of RecordInstance
        """
        data = values.of(
            {
                "Category": category,
                "StartDate": serialize.iso8601_date(start_date),
                "EndDate": serialize.iso8601_date(end_date),
                "IncludeSubaccounts": serialize.boolean_to_string(include_subaccounts),
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return RecordPage(self._version, response, self._solution)

    async def page_async(
        self,
        category: Union[str, object] = values.unset,
        start_date: Union[date, object] = values.unset,
        end_date: Union[date, object] = values.unset,
        include_subaccounts: Union[bool, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> RecordPage:
        """
        Asynchronously retrieve a single page of RecordInstance records from the API.
        Request is executed immediately

        :param category: The [usage category](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) of the UsageRecord resources to read. Only UsageRecord resources in the specified category are retrieved.
        :param start_date: Only include usage that has occurred on or after this date. Specify the date in GMT and format as `YYYY-MM-DD`. You can also specify offsets from the current date, such as: `-30days`, which will set the start date to be 30 days before the current date.
        :param end_date: Only include usage that occurred on or before this date. Specify the date in GMT and format as `YYYY-MM-DD`.  You can also specify offsets from the current date, such as: `+30days`, which will set the end date to 30 days from the current date.
        :param include_subaccounts: Whether to include usage from the master account and all its subaccounts. Can be: `true` (the default) to include usage from the master account and all subaccounts or `false` to retrieve usage from only the specified account.
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of RecordInstance
        """
        data = values.of(
            {
                "Category": category,
                "StartDate": serialize.iso8601_date(start_date),
                "EndDate": serialize.iso8601_date(end_date),
                "IncludeSubaccounts": serialize.boolean_to_string(include_subaccounts),
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return RecordPage(self._version, response, self._solution)

    def get_page(self, target_url: str) -> RecordPage:
        """
        Retrieve a specific page of RecordInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of RecordInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return RecordPage(self._version, response, self._solution)

    async def get_page_async(self, target_url: str) -> RecordPage:
        """
        Asynchronously retrieve a specific page of RecordInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of RecordInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return RecordPage(self._version, response, self._solution)

    @property
    def all_time(self) -> AllTimeList:
        """
        Access the all_time
        """
        if self._all_time is None:
            self._all_time = AllTimeList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._all_time

    @property
    def daily(self) -> DailyList:
        """
        Access the daily
        """
        if self._daily is None:
            self._daily = DailyList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._daily

    @property
    def last_month(self) -> LastMonthList:
        """
        Access the last_month
        """
        if self._last_month is None:
            self._last_month = LastMonthList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._last_month

    @property
    def monthly(self) -> MonthlyList:
        """
        Access the monthly
        """
        if self._monthly is None:
            self._monthly = MonthlyList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._monthly

    @property
    def this_month(self) -> ThisMonthList:
        """
        Access the this_month
        """
        if self._this_month is None:
            self._this_month = ThisMonthList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._this_month

    @property
    def today(self) -> TodayList:
        """
        Access the today
        """
        if self._today is None:
            self._today = TodayList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._today

    @property
    def yearly(self) -> YearlyList:
        """
        Access the yearly
        """
        if self._yearly is None:
            self._yearly = YearlyList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._yearly

    @property
    def yesterday(self) -> YesterdayList:
        """
        Access the yesterday
        """
        if self._yesterday is None:
            self._yesterday = YesterdayList(
                self._version, account_sid=self._solution["account_sid"]
            )
        return self._yesterday

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.RecordList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Lookups
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, List, Optional, Union
from twilio.base import values

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class RateLimitInstance(InstanceResource):
    """
    :ivar rate_limits:
    """

    def __init__(self, version: Version, payload: Dict[str, Any]):
        super().__init__(version)

        self.rate_limits: Optional[List[str]] = payload.get("rate_limits")

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """

        return "<Twilio.Lookups.V2.RateLimitInstance>"


class RateLimitList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the RateLimitList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/RateLimits"

    def fetch(
        self, fields: Union[List[str], object] = values.unset
    ) -> RateLimitInstance:
        """
        Asynchronously fetch the RateLimitInstance

        :param fields:
        :returns: The fetched RateLimitInstance
        """
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        params = values.of(
            {
                "Fields": fields,
            }
        )

        payload = self._version.fetch(
            method="GET", uri=self._uri, headers=headers, params=params
        )

        return RateLimitInstance(self._version, payload)

    async def fetch_async(
        self, fields: Union[List[str], object] = values.unset
    ) -> RateLimitInstance:
        """
        Asynchronously fetch the RateLimitInstance

        :param fields:
        :returns: The fetched RateLimitInstance
        """
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        params = values.of(
            {
                "Fields": fields,
            }
        )

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers, params=params
        )

        return RateLimitInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Lookups.V2.RateLimitList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Chat
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Optional
from twilio.base.version import Version
from twilio.base.domain import Domain
from twilio.rest.chat.v2.credential import CredentialList
from twilio.rest.chat.v2.service import ServiceList


class V2(Version):

    def __init__(self, domain: Domain):
        """
        Initialize the V2 version of Chat

        :param domain: The Twilio.chat domain
        """
        super().__init__(domain, "v2")
        self._credentials: Optional[CredentialList] = None
        self._services: Optional[ServiceList] = None

    @property
    def credentials(self) -> CredentialList:
        if self._credentials is None:
            self._credentials = CredentialList(self)
        return self._credentials

    @property
    def services(self) -> ServiceList:
        if self._services is None:
            self._services = ServiceList(self)
        return self._services

    def __repr__(self) -> str:
        """
        Provide a friendly representation
        :returns: Machine friendly representation
        """
        return "<Twilio.Chat.V2>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Api
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class TriggerInstance(InstanceResource):

    class Recurring(object):
        DAILY = "daily"
        MONTHLY = "monthly"
        YEARLY = "yearly"
        ALLTIME = "alltime"

    class TriggerField(object):
        COUNT = "count"
        USAGE = "usage"
        PRICE = "price"

    """
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that the trigger monitors.
    :ivar api_version: The API version used to create the resource.
    :ivar callback_method: The HTTP method we use to call `callback_url`. Can be: `GET` or `POST`.
    :ivar callback_url: The URL we call using the `callback_method` when the trigger fires.
    :ivar current_value: The current value of the field the trigger is watching.
    :ivar date_created: The date and time in GMT that the resource was created specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar date_fired: The date and time in GMT that the trigger was last fired specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar date_updated: The date and time in GMT that the resource was last updated specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
    :ivar friendly_name: The string that you assigned to describe the trigger.
    :ivar recurring: 
    :ivar sid: The unique string that that we created to identify the UsageTrigger resource.
    :ivar trigger_by: 
    :ivar trigger_value: The value at which the trigger will fire.  Must be a positive, numeric value.
    :ivar uri: The URI of the resource, relative to `https://api.twilio.com`.
    :ivar usage_category: The usage category the trigger watches. Must be one of the supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
    :ivar usage_record_uri: The URI of the [UsageRecord](https://www.twilio.com/docs/usage/api/usage-record) resource this trigger watches, relative to `https://api.twilio.com`.
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        account_sid: str,
        sid: Optional[str] = None,
    ):
        super().__init__(version)

        self.account_sid: Optional[str] = payload.get("account_sid")
        self.api_version: Optional[str] = payload.get("api_version")
        self.callback_method: Optional[str] = payload.get("callback_method")
        self.callback_url: Optional[str] = payload.get("callback_url")
        self.current_value: Optional[str] = payload.get("current_value")
        self.date_created: Optional[datetime] = deserialize.rfc2822_datetime(
            payload.get("date_created")
        )
        self.date_fired: Optional[datetime] = deserialize.rfc2822_datetime(
            payload.get("date_fired")
        )
        self.date_updated: Optional[datetime] = deserialize.rfc2822_datetime(
            payload.get("date_updated")
        )
        self.friendly_name: Optional[str] = payload.get("friendly_name")
        self.recurring: Optional["TriggerInstance.Recurring"] = payload.get("recurring")
        self.sid: Optional[str] = payload.get("sid")
        self.trigger_by: Optional["TriggerInstance.TriggerField"] = payload.get(
            "trigger_by"
        )
        self.trigger_value: Optional[str] = payload.get("trigger_value")
        self.uri: Optional[str] = payload.get("uri")
        self.usage_category: Optional[str] = payload.get("usage_category")
        self.usage_record_uri: Optional[str] = payload.get("usage_record_uri")

        self._solution = {
            "account_sid": account_sid,
            "sid": sid or self.sid,
        }
        self._context: Optional[TriggerContext] = None

    @property
    def _proxy(self) -> "TriggerContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: TriggerContext for this TriggerInstance
        """
        if self._context is None:
            self._context = TriggerContext(
                self._version,
                account_sid=self._solution["account_sid"],
                sid=self._solution["sid"],
            )
        return self._context

    def delete(self) -> bool:
        """
        Deletes the TriggerInstance


        :returns: True if delete succeeds, False otherwise
        """
        return self._proxy.delete()

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the TriggerInstance


        :returns: True if delete succeeds, False otherwise
        """
        return await self._proxy.delete_async()

    def fetch(self) -> "TriggerInstance":
        """
        Fetch the TriggerInstance


        :returns: The fetched TriggerInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "TriggerInstance":
        """
        Asynchronous coroutine to fetch the TriggerInstance


        :returns: The fetched TriggerInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self,
        callback_method: Union[str, object] = values.unset,
        callback_url: Union[str, object] = values.unset,
        friendly_name: Union[str, object] = values.unset,
    ) -> "TriggerInstance":
        """
        Update the TriggerInstance

        :param callback_method: The HTTP method we should use to call `callback_url`. Can be: `GET` or `POST` and the default is `POST`.
        :param callback_url: The URL we should call using `callback_method` when the trigger fires.
        :param friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.

        :returns: The updated TriggerInstance
        """
        return self._proxy.update(
            callback_method=callback_method,
            callback_url=callback_url,
            friendly_name=friendly_name,
        )

    async def update_async(
        self,
        callback_method: Union[str, object] = values.unset,
        callback_url: Union[str, object] = values.unset,
        friendly_name: Union[str, object] = values.unset,
    ) -> "TriggerInstance":
        """
        Asynchronous coroutine to update the TriggerInstance

        :param callback_method: The HTTP method we should use to call `callback_url`. Can be: `GET` or `POST` and the default is `POST`.
        :param callback_url: The URL we should call using `callback_method` when the trigger fires.
        :param friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.

        :returns: The updated TriggerInstance
        """
        return await self._proxy.update_async(
            callback_method=callback_method,
            callback_url=callback_url,
            friendly_name=friendly_name,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.TriggerInstance {}>".format(context)


class TriggerContext(InstanceContext):

    def __init__(self, version: Version, account_sid: str, sid: str):
        """
        Initialize the TriggerContext

        :param version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the UsageTrigger resources to update.
        :param sid: The Twilio-provided string that uniquely identifies the UsageTrigger resource to update.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
            "sid": sid,
        }
        self._uri = "/Accounts/{account_sid}/Usage/Triggers/{sid}.json".format(
            **self._solution
        )

    def delete(self) -> bool:
        """
        Deletes the TriggerInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        return self._version.delete(method="DELETE", uri=self._uri, headers=headers)

    async def delete_async(self) -> bool:
        """
        Asynchronous coroutine that deletes the TriggerInstance


        :returns: True if delete succeeds, False otherwise
        """

        headers = values.of({})

        return await self._version.delete_async(
            method="DELETE", uri=self._uri, headers=headers
        )

    def fetch(self) -> TriggerInstance:
        """
        Fetch the TriggerInstance


        :returns: The fetched TriggerInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return TriggerInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    async def fetch_async(self) -> TriggerInstance:
        """
        Asynchronous coroutine to fetch the TriggerInstance


        :returns: The fetched TriggerInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return TriggerInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    def update(
        self,
        callback_method: Union[str, object] = values.unset,
        callback_url: Union[str, object] = values.unset,
        friendly_name: Union[str, object] = values.unset,
    ) -> TriggerInstance:
        """
        Update the TriggerInstance

        :param callback_method: The HTTP method we should use to call `callback_url`. Can be: `GET` or `POST` and the default is `POST`.
        :param callback_url: The URL we should call using `callback_method` when the trigger fires.
        :param friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.

        :returns: The updated TriggerInstance
        """

        data = values.of(
            {
                "CallbackMethod": callback_method,
                "CallbackUrl": callback_url,
                "FriendlyName": friendly_name,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return TriggerInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    async def update_async(
        self,
        callback_method: Union[str, object] = values.unset,
        callback_url: Union[str, object] = values.unset,
        friendly_name: Union[str, object] = values.unset,
    ) -> TriggerInstance:
        """
        Asynchronous coroutine to update the TriggerInstance

        :param callback_method: The HTTP method we should use to call `callback_url`. Can be: `GET` or `POST` and the default is `POST`.
        :param callback_url: The URL we should call using `callback_method` when the trigger fires.
        :param friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.

        :returns: The updated TriggerInstance
        """

        data = values.of(
            {
                "CallbackMethod": callback_method,
                "CallbackUrl": callback_url,
                "FriendlyName": friendly_name,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return TriggerInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.TriggerContext {}>".format(context)


class TriggerPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> TriggerInstance:
        """
        Build an instance of TriggerInstance

        :param payload: Payload response from the API
        """
        return TriggerInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.TriggerPage>"


class TriggerList(ListResource):

    def __init__(self, version: Version, account_sid: str):
        """
        Initialize the TriggerList

        :param version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the UsageTrigger resources to read.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
        }
        self._uri = "/Accounts/{account_sid}/Usage/Triggers.json".format(
            **self._solution
        )

    def create(
        self,
        callback_url: str,
        trigger_value: str,
        usage_category: str,
        callback_method: Union[str, object] = values.unset,
        friendly_name: Union[str, object] = values.unset,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
    ) -> TriggerInstance:
        """
        Create the TriggerInstance

        :param callback_url: The URL we should call using `callback_method` when the trigger fires.
        :param trigger_value: The usage value at which the trigger should fire.  For convenience, you can use an offset value such as `+30` to specify a trigger_value that is 30 units more than the current usage value. Be sure to urlencode a `+` as `%2B`.
        :param usage_category: The usage category that the trigger should watch.  Use one of the supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) for this value.
        :param callback_method: The HTTP method we should use to call `callback_url`. Can be: `GET` or `POST` and the default is `POST`.
        :param friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.
        :param recurring:
        :param trigger_by:

        :returns: The created TriggerInstance
        """

        data = values.of(
            {
                "CallbackUrl": callback_url,
                "TriggerValue": trigger_value,
                "UsageCategory": usage_category,
                "CallbackMethod": callback_method,
                "FriendlyName": friendly_name,
                "Recurring": recurring,
                "TriggerBy": trigger_by,
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return TriggerInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    async def create_async(
        self,
        callback_url: str,
        trigger_value: str,
        usage_category: str,
        callback_method: Union[str, object] = values.unset,
        friendly_name: Union[str, object] = values.unset,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
    ) -> TriggerInstance:
        """
        Asynchronously create the TriggerInstance

        :param callback_url: The URL we should call using `callback_method` when the trigger fires.
        :param trigger_value: The usage value at which the trigger should fire.  For convenience, you can use an offset value such as `+30` to specify a trigger_value that is 30 units more than the current usage value. Be sure to urlencode a `+` as `%2B`.
        :param usage_category: The usage category that the trigger should watch.  Use one of the supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories) for this value.
        :param callback_method: The HTTP method we should use to call `callback_url`. Can be: `GET` or `POST` and the default is `POST`.
        :param friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.
        :param recurring:
        :param trigger_by:

        :returns: The created TriggerInstance
        """

        data = values.of(
            {
                "CallbackUrl": callback_url,
                "TriggerValue": trigger_value,
                "UsageCategory": usage_category,
                "CallbackMethod": callback_method,
                "FriendlyName": friendly_name,
                "Recurring": recurring,
                "TriggerBy": trigger_by,
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return TriggerInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    def stream(
        self,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
        usage_category: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[TriggerInstance]:
        """
        Streams TriggerInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param &quot;TriggerInstance.Recurring&quot; recurring: The frequency of recurring UsageTriggers to read. Can be: `daily`, `monthly`, or `yearly` to read recurring UsageTriggers. An empty value or a value of `alltime` reads non-recurring UsageTriggers.
        :param &quot;TriggerInstance.TriggerField&quot; trigger_by: The trigger field of the UsageTriggers to read.  Can be: `count`, `usage`, or `price` as described in the [UsageRecords documentation](https://www.twilio.com/docs/usage/api/usage-record#usage-count-price).
        :param str usage_category: The usage category of the UsageTriggers to read. Must be a supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(
            recurring=recurring,
            trigger_by=trigger_by,
            usage_category=usage_category,
            page_size=limits["page_size"],
        )

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
        usage_category: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[TriggerInstance]:
        """
        Asynchronously streams TriggerInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param &quot;TriggerInstance.Recurring&quot; recurring: The frequency of recurring UsageTriggers to read. Can be: `daily`, `monthly`, or `yearly` to read recurring UsageTriggers. An empty value or a value of `alltime` reads non-recurring UsageTriggers.
        :param &quot;TriggerInstance.TriggerField&quot; trigger_by: The trigger field of the UsageTriggers to read.  Can be: `count`, `usage`, or `price` as described in the [UsageRecords documentation](https://www.twilio.com/docs/usage/api/usage-record#usage-count-price).
        :param str usage_category: The usage category of the UsageTriggers to read. Must be a supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            recurring=recurring,
            trigger_by=trigger_by,
            usage_category=usage_category,
            page_size=limits["page_size"],
        )

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
        usage_category: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[TriggerInstance]:
        """
        Lists TriggerInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param &quot;TriggerInstance.Recurring&quot; recurring: The frequency of recurring UsageTriggers to read. Can be: `daily`, `monthly`, or `yearly` to read recurring UsageTriggers. An empty value or a value of `alltime` reads non-recurring UsageTriggers.
        :param &quot;TriggerInstance.TriggerField&quot; trigger_by: The trigger field of the UsageTriggers to read.  Can be: `count`, `usage`, or `price` as described in the [UsageRecords documentation](https://www.twilio.com/docs/usage/api/usage-record#usage-count-price).
        :param str usage_category: The usage category of the UsageTriggers to read. Must be a supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                recurring=recurring,
                trigger_by=trigger_by,
                usage_category=usage_category,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
        usage_category: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[TriggerInstance]:
        """
        Asynchronously lists TriggerInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param &quot;TriggerInstance.Recurring&quot; recurring: The frequency of recurring UsageTriggers to read. Can be: `daily`, `monthly`, or `yearly` to read recurring UsageTriggers. An empty value or a value of `alltime` reads non-recurring UsageTriggers.
        :param &quot;TriggerInstance.TriggerField&quot; trigger_by: The trigger field of the UsageTriggers to read.  Can be: `count`, `usage`, or `price` as described in the [UsageRecords documentation](https://www.twilio.com/docs/usage/api/usage-record#usage-count-price).
        :param str usage_category: The usage category of the UsageTriggers to read. Must be a supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                recurring=recurring,
                trigger_by=trigger_by,
                usage_category=usage_category,
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
        usage_category: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> TriggerPage:
        """
        Retrieve a single page of TriggerInstance records from the API.
        Request is executed immediately

        :param recurring: The frequency of recurring UsageTriggers to read. Can be: `daily`, `monthly`, or `yearly` to read recurring UsageTriggers. An empty value or a value of `alltime` reads non-recurring UsageTriggers.
        :param trigger_by: The trigger field of the UsageTriggers to read.  Can be: `count`, `usage`, or `price` as described in the [UsageRecords documentation](https://www.twilio.com/docs/usage/api/usage-record#usage-count-price).
        :param usage_category: The usage category of the UsageTriggers to read. Must be a supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of TriggerInstance
        """
        data = values.of(
            {
                "Recurring": recurring,
                "TriggerBy": trigger_by,
                "UsageCategory": usage_category,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return TriggerPage(self._version, response, self._solution)

    async def page_async(
        self,
        recurring: Union["TriggerInstance.Recurring", object] = values.unset,
        trigger_by: Union["TriggerInstance.TriggerField", object] = values.unset,
        usage_category: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> TriggerPage:
        """
        Asynchronously retrieve a single page of TriggerInstance records from the API.
        Request is executed immediately

        :param recurring: The frequency of recurring UsageTriggers to read. Can be: `daily`, `monthly`, or `yearly` to read recurring UsageTriggers. An empty value or a value of `alltime` reads non-recurring UsageTriggers.
        :param trigger_by: The trigger field of the UsageTriggers to read.  Can be: `count`, `usage`, or `price` as described in the [UsageRecords documentation](https://www.twilio.com/docs/usage/api/usage-record#usage-count-price).
        :param usage_category: The usage category of the UsageTriggers to read. Must be a supported [usage categories](https://www.twilio.com/docs/usage/api/usage-record#usage-categories).
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of TriggerInstance
        """
        data = values.of(
            {
                "Recurring": recurring,
                "TriggerBy": trigger_by,
                "UsageCategory": usage_category,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return TriggerPage(self._version, response, self._solution)

    def get_page(self, target_url: str) -> TriggerPage:
        """
        Retrieve a specific page of TriggerInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of TriggerInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return TriggerPage(self._version, response, self._solution)

    async def get_page_async(self, target_url: str) -> TriggerPage:
        """
        Asynchronously retrieve a specific page of TriggerInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of TriggerInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return TriggerPage(self._version, response, self._solution)

    def get(self, sid: str) -> TriggerContext:
        """
        Constructs a TriggerContext

        :param sid: The Twilio-provided string that uniquely identifies the UsageTrigger resource to update.
        """
        return TriggerContext(
            self._version, account_sid=self._solution["account_sid"], sid=sid
        )

    def __call__(self, sid: str) -> TriggerContext:
        """
        Constructs a TriggerContext

        :param sid: The Twilio-provided string that uniquely identifies the UsageTrigger resource to update.
        """
        return TriggerContext(
            self._version, account_sid=self._solution["account_sid"], sid=sid
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.TriggerList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Lookups
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from twilio.base import values

from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class QueryInstance(InstanceResource):

    class CallForwarding(object):
        """
        :ivar call_forwarding_enabled:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.call_forwarding_enabled: Optional[bool] = payload.get(
                "call_forwarding_enabled"
            )
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.call_forwarding_enabled,
                "": self.error_code,
            }

    class CallerName(object):
        """
        :ivar caller_name:
        :ivar caller_type:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.caller_name: Optional[str] = payload.get("caller_name")
            self.caller_type: Optional[str] = payload.get("caller_type")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.caller_name,
                "": self.caller_type,
                "": self.error_code,
            }

    class IdentityMatch(object):
        """
        :ivar first_name_match:
        :ivar last_name_match:
        :ivar address_line_match:
        :ivar city_match:
        :ivar state_match:
        :ivar postal_code_match:
        :ivar country_code_match:
        :ivar national_id_match:
        :ivar date_of_birth_match:
        :ivar summary_score:
        :ivar error_code:
        :ivar error_message:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.first_name_match: Optional[str] = payload.get("first_name_match")
            self.last_name_match: Optional[str] = payload.get("last_name_match")
            self.address_line_match: Optional[str] = payload.get("address_line_match")
            self.city_match: Optional[str] = payload.get("city_match")
            self.state_match: Optional[str] = payload.get("state_match")
            self.postal_code_match: Optional[str] = payload.get("postal_code_match")
            self.country_code_match: Optional[str] = payload.get("country_code_match")
            self.national_id_match: Optional[str] = payload.get("national_id_match")
            self.date_of_birth_match: Optional[str] = payload.get("date_of_birth_match")
            self.summary_score: Optional[int] = payload.get("summary_score")
            self.error_code: Optional[int] = payload.get("error_code")
            self.error_message: Optional[str] = payload.get("error_message")

        def to_dict(self):
            return {
                "": self.first_name_match,
                "": self.last_name_match,
                "": self.address_line_match,
                "": self.city_match,
                "": self.state_match,
                "": self.postal_code_match,
                "": self.country_code_match,
                "": self.national_id_match,
                "": self.date_of_birth_match,
                "": self.summary_score,
                "": self.error_code,
                "": self.error_message,
            }

    class IdentityMatchParameters(object):
        """
        :ivar first_name:
        :ivar last_name:
        :ivar address_line1:
        :ivar address_line2:
        :ivar city:
        :ivar state:
        :ivar postal_code:
        :ivar address_country_code:
        :ivar national_id:
        :ivar date_of_birth:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.first_name: Optional[str] = payload.get("first_name")
            self.last_name: Optional[str] = payload.get("last_name")
            self.address_line1: Optional[str] = payload.get("address_line1")
            self.address_line2: Optional[str] = payload.get("address_line2")
            self.city: Optional[str] = payload.get("city")
            self.state: Optional[str] = payload.get("state")
            self.postal_code: Optional[str] = payload.get("postal_code")
            self.address_country_code: Optional[str] = payload.get(
                "address_country_code"
            )
            self.national_id: Optional[str] = payload.get("national_id")
            self.date_of_birth: Optional[str] = payload.get("date_of_birth")

        def to_dict(self):
            return {
                "first_name": self.first_name,
                "last_name": self.last_name,
                "address_line1": self.address_line1,
                "address_line2": self.address_line2,
                "city": self.city,
                "state": self.state,
                "postal_code": self.postal_code,
                "address_country_code": self.address_country_code,
                "national_id": self.national_id,
                "date_of_birth": self.date_of_birth,
            }

    class LastSimSwap(object):
        """
        :ivar last_sim_swap_date:
        :ivar swapped_period:
        :ivar swapped_in_period:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_sim_swap_date: Optional[datetime] = payload.get(
                "last_sim_swap_date"
            )
            self.swapped_period: Optional[str] = payload.get("swapped_period")
            self.swapped_in_period: Optional[bool] = payload.get("swapped_in_period")

        def to_dict(self):
            return {
                "": self.last_sim_swap_date,
                "": self.swapped_period,
                "": self.swapped_in_period,
            }

    class LineStatus(object):
        """
        :ivar status:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.status: Optional[str] = payload.get("status")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.status,
                "": self.error_code,
            }

    class LineTypeIntelligence(object):
        """
        :ivar type:
        :ivar carrier_name:
        :ivar mobile_country_code:
        :ivar mobile_network_code:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.carrier_name: Optional[str] = payload.get("carrier_name")
            self.mobile_country_code: Optional[str] = payload.get("mobile_country_code")
            self.mobile_network_code: Optional[str] = payload.get("mobile_network_code")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.type,
                "": self.carrier_name,
                "": self.mobile_country_code,
                "": self.mobile_network_code,
                "": self.error_code,
            }

    class LookupRequest1(object):
        """
        :ivar phone_numbers:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.phone_numbers: Optional[List[QueryList.LookupRequestWithCorId]] = (
                payload.get("phone_numbers")
            )

        def to_dict(self):
            return {
                "phone_numbers": (
                    [phone_numbers.to_dict() for phone_numbers in self.phone_numbers]
                    if self.phone_numbers is not None
                    else None
                ),
            }

    class LookupRequestWithCorId(object):
        """
        :ivar correlation_id: Unique identifier used to match request with response
        :ivar phone_number:
        :ivar fields:
        :ivar country_code:
        :ivar identity_match:
        :ivar reassigned_number:
        :ivar sms_pumping_risk:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.correlation_id: Optional[str] = payload.get("correlation_id")
            self.phone_number: Optional[str] = payload.get("phone_number")
            self.fields: Optional[List[Enumstr]] = payload.get("fields")
            self.country_code: Optional[str] = payload.get("country_code")
            self.identity_match: Optional[QueryList.IdentityMatchParameters] = (
                payload.get("identity_match")
            )
            self.reassigned_number: Optional[QueryList.ReassignedNumberRequest] = (
                payload.get("reassigned_number")
            )
            self.sms_pumping_risk: Optional[QueryList.SmsPumpingRiskParameters] = (
                payload.get("sms_pumping_risk")
            )

        def to_dict(self):
            return {
                "correlation_id": self.correlation_id,
                "phone_number": self.phone_number,
                "fields": self.fields,
                "country_code": self.country_code,
                "identity_match": (
                    self.identity_match.to_dict()
                    if self.identity_match is not None
                    else None
                ),
                "reassigned_number": (
                    self.reassigned_number.to_dict()
                    if self.reassigned_number is not None
                    else None
                ),
                "sms_pumping_risk": (
                    self.sms_pumping_risk.to_dict()
                    if self.sms_pumping_risk is not None
                    else None
                ),
            }

    class ReassignedNumberRequest(object):
        """
        :ivar last_verified_date:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_verified_date: Optional[str] = payload.get("last_verified_date")

        def to_dict(self):
            return {
                "last_verified_date": self.last_verified_date,
            }

    class ReassignedNumberResponse(object):
        """
        :ivar last_verified_date:
        :ivar is_number_reassigned:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_verified_date: Optional[str] = payload.get("last_verified_date")
            self.is_number_reassigned: Optional[str] = payload.get(
                "is_number_reassigned"
            )
            self.error_code: Optional[str] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.last_verified_date,
                "": self.is_number_reassigned,
                "": self.error_code,
            }

    class SimSwap(object):
        """
        :ivar last_sim_swap:
        :ivar carrier_name:
        :ivar mobile_country_code:
        :ivar mobile_network_code:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_sim_swap: Optional[LastSimSwap] = payload.get("last_sim_swap")
            self.carrier_name: Optional[str] = payload.get("carrier_name")
            self.mobile_country_code: Optional[str] = payload.get("mobile_country_code")
            self.mobile_network_code: Optional[str] = payload.get("mobile_network_code")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": (
                    self.last_sim_swap.to_dict()
                    if self.last_sim_swap is not None
                    else None
                ),
                "": self.carrier_name,
                "": self.mobile_country_code,
                "": self.mobile_network_code,
                "": self.error_code,
            }

    class SmsPumpingRisk(object):
        """
        :ivar carrier_risk_category:
        :ivar number_blocked:
        :ivar number_blocked_date:
        :ivar number_blocked_last_3_months:
        :ivar sms_pumping_risk_score:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.carrier_risk_category: Optional[str] = payload.get(
                "carrier_risk_category"
            )
            self.number_blocked: Optional[bool] = payload.get("number_blocked")
            self.number_blocked_date: Optional[datetime] = payload.get(
                "number_blocked_date"
            )
            self.number_blocked_last_3_months: Optional[bool] = payload.get(
                "number_blocked_last_3_months"
            )
            self.sms_pumping_risk_score: Optional[int] = payload.get(
                "sms_pumping_risk_score"
            )
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.carrier_risk_category,
                "": self.number_blocked,
                "": self.number_blocked_date,
                "": self.number_blocked_last_3_months,
                "": self.sms_pumping_risk_score,
                "": self.error_code,
            }

    class SmsPumpingRiskParameters(object):
        """
        :ivar partner_sub_id:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.partner_sub_id: Optional[str] = payload.get("partner_sub_id")

        def to_dict(self):
            return {
                "partner_sub_id": self.partner_sub_id,
            }

    """
    :ivar phone_numbers: 
    """

    def __init__(self, version: Version, payload: Dict[str, Any]):
        super().__init__(version)

        self.phone_numbers: Optional[List[str]] = payload.get("phone_numbers")

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """

        return "<Twilio.Lookups.V2.QueryInstance>"


class QueryList(ListResource):

    class CallForwarding(object):
        """
        :ivar call_forwarding_enabled:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.call_forwarding_enabled: Optional[bool] = payload.get(
                "call_forwarding_enabled"
            )
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.call_forwarding_enabled,
                "": self.error_code,
            }

    class CallerName(object):
        """
        :ivar caller_name:
        :ivar caller_type:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.caller_name: Optional[str] = payload.get("caller_name")
            self.caller_type: Optional[str] = payload.get("caller_type")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.caller_name,
                "": self.caller_type,
                "": self.error_code,
            }

    class IdentityMatch(object):
        """
        :ivar first_name_match:
        :ivar last_name_match:
        :ivar address_line_match:
        :ivar city_match:
        :ivar state_match:
        :ivar postal_code_match:
        :ivar country_code_match:
        :ivar national_id_match:
        :ivar date_of_birth_match:
        :ivar summary_score:
        :ivar error_code:
        :ivar error_message:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.first_name_match: Optional[str] = payload.get("first_name_match")
            self.last_name_match: Optional[str] = payload.get("last_name_match")
            self.address_line_match: Optional[str] = payload.get("address_line_match")
            self.city_match: Optional[str] = payload.get("city_match")
            self.state_match: Optional[str] = payload.get("state_match")
            self.postal_code_match: Optional[str] = payload.get("postal_code_match")
            self.country_code_match: Optional[str] = payload.get("country_code_match")
            self.national_id_match: Optional[str] = payload.get("national_id_match")
            self.date_of_birth_match: Optional[str] = payload.get("date_of_birth_match")
            self.summary_score: Optional[int] = payload.get("summary_score")
            self.error_code: Optional[int] = payload.get("error_code")
            self.error_message: Optional[str] = payload.get("error_message")

        def to_dict(self):
            return {
                "": self.first_name_match,
                "": self.last_name_match,
                "": self.address_line_match,
                "": self.city_match,
                "": self.state_match,
                "": self.postal_code_match,
                "": self.country_code_match,
                "": self.national_id_match,
                "": self.date_of_birth_match,
                "": self.summary_score,
                "": self.error_code,
                "": self.error_message,
            }

    class IdentityMatchParameters(object):
        """
        :ivar first_name:
        :ivar last_name:
        :ivar address_line1:
        :ivar address_line2:
        :ivar city:
        :ivar state:
        :ivar postal_code:
        :ivar address_country_code:
        :ivar national_id:
        :ivar date_of_birth:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.first_name: Optional[str] = payload.get("first_name")
            self.last_name: Optional[str] = payload.get("last_name")
            self.address_line1: Optional[str] = payload.get("address_line1")
            self.address_line2: Optional[str] = payload.get("address_line2")
            self.city: Optional[str] = payload.get("city")
            self.state: Optional[str] = payload.get("state")
            self.postal_code: Optional[str] = payload.get("postal_code")
            self.address_country_code: Optional[str] = payload.get(
                "address_country_code"
            )
            self.national_id: Optional[str] = payload.get("national_id")
            self.date_of_birth: Optional[str] = payload.get("date_of_birth")

        def to_dict(self):
            return {
                "first_name": self.first_name,
                "last_name": self.last_name,
                "address_line1": self.address_line1,
                "address_line2": self.address_line2,
                "city": self.city,
                "state": self.state,
                "postal_code": self.postal_code,
                "address_country_code": self.address_country_code,
                "national_id": self.national_id,
                "date_of_birth": self.date_of_birth,
            }

    class LastSimSwap(object):
        """
        :ivar last_sim_swap_date:
        :ivar swapped_period:
        :ivar swapped_in_period:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_sim_swap_date: Optional[datetime] = payload.get(
                "last_sim_swap_date"
            )
            self.swapped_period: Optional[str] = payload.get("swapped_period")
            self.swapped_in_period: Optional[bool] = payload.get("swapped_in_period")

        def to_dict(self):
            return {
                "": self.last_sim_swap_date,
                "": self.swapped_period,
                "": self.swapped_in_period,
            }

    class LineStatus(object):
        """
        :ivar status:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.status: Optional[str] = payload.get("status")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.status,
                "": self.error_code,
            }

    class LineTypeIntelligence(object):
        """
        :ivar type:
        :ivar carrier_name:
        :ivar mobile_country_code:
        :ivar mobile_network_code:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.type: Optional[str] = payload.get("type")
            self.carrier_name: Optional[str] = payload.get("carrier_name")
            self.mobile_country_code: Optional[str] = payload.get("mobile_country_code")
            self.mobile_network_code: Optional[str] = payload.get("mobile_network_code")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.type,
                "": self.carrier_name,
                "": self.mobile_country_code,
                "": self.mobile_network_code,
                "": self.error_code,
            }

    class LookupRequest1(object):
        """
        :ivar phone_numbers:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.phone_numbers: Optional[List[QueryList.LookupRequestWithCorId]] = (
                payload.get("phone_numbers")
            )

        def to_dict(self):
            return {
                "phone_numbers": (
                    [phone_numbers.to_dict() for phone_numbers in self.phone_numbers]
                    if self.phone_numbers is not None
                    else None
                ),
            }

    class LookupRequestWithCorId(object):
        """
        :ivar correlation_id: Unique identifier used to match request with response
        :ivar phone_number:
        :ivar fields:
        :ivar country_code:
        :ivar identity_match:
        :ivar reassigned_number:
        :ivar sms_pumping_risk:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.correlation_id: Optional[str] = payload.get("correlation_id")
            self.phone_number: Optional[str] = payload.get("phone_number")
            self.fields: Optional[List[Enumstr]] = payload.get("fields")
            self.country_code: Optional[str] = payload.get("country_code")
            self.identity_match: Optional[QueryList.IdentityMatchParameters] = (
                payload.get("identity_match")
            )
            self.reassigned_number: Optional[QueryList.ReassignedNumberRequest] = (
                payload.get("reassigned_number")
            )
            self.sms_pumping_risk: Optional[QueryList.SmsPumpingRiskParameters] = (
                payload.get("sms_pumping_risk")
            )

        def to_dict(self):
            return {
                "correlation_id": self.correlation_id,
                "phone_number": self.phone_number,
                "fields": self.fields,
                "country_code": self.country_code,
                "identity_match": (
                    self.identity_match.to_dict()
                    if self.identity_match is not None
                    else None
                ),
                "reassigned_number": (
                    self.reassigned_number.to_dict()
                    if self.reassigned_number is not None
                    else None
                ),
                "sms_pumping_risk": (
                    self.sms_pumping_risk.to_dict()
                    if self.sms_pumping_risk is not None
                    else None
                ),
            }

    class ReassignedNumberRequest(object):
        """
        :ivar last_verified_date:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_verified_date: Optional[str] = payload.get("last_verified_date")

        def to_dict(self):
            return {
                "last_verified_date": self.last_verified_date,
            }

    class ReassignedNumberResponse(object):
        """
        :ivar last_verified_date:
        :ivar is_number_reassigned:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_verified_date: Optional[str] = payload.get("last_verified_date")
            self.is_number_reassigned: Optional[str] = payload.get(
                "is_number_reassigned"
            )
            self.error_code: Optional[str] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.last_verified_date,
                "": self.is_number_reassigned,
                "": self.error_code,
            }

    class SimSwap(object):
        """
        :ivar last_sim_swap:
        :ivar carrier_name:
        :ivar mobile_country_code:
        :ivar mobile_network_code:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.last_sim_swap: Optional[LastSimSwap] = payload.get("last_sim_swap")
            self.carrier_name: Optional[str] = payload.get("carrier_name")
            self.mobile_country_code: Optional[str] = payload.get("mobile_country_code")
            self.mobile_network_code: Optional[str] = payload.get("mobile_network_code")
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": (
                    self.last_sim_swap.to_dict()
                    if self.last_sim_swap is not None
                    else None
                ),
                "": self.carrier_name,
                "": self.mobile_country_code,
                "": self.mobile_network_code,
                "": self.error_code,
            }

    class SmsPumpingRisk(object):
        """
        :ivar carrier_risk_category:
        :ivar number_blocked:
        :ivar number_blocked_date:
        :ivar number_blocked_last_3_months:
        :ivar sms_pumping_risk_score:
        :ivar error_code:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.carrier_risk_category: Optional[str] = payload.get(
                "carrier_risk_category"
            )
            self.number_blocked: Optional[bool] = payload.get("number_blocked")
            self.number_blocked_date: Optional[datetime] = payload.get(
                "number_blocked_date"
            )
            self.number_blocked_last_3_months: Optional[bool] = payload.get(
                "number_blocked_last_3_months"
            )
            self.sms_pumping_risk_score: Optional[int] = payload.get(
                "sms_pumping_risk_score"
            )
            self.error_code: Optional[int] = payload.get("error_code")

        def to_dict(self):
            return {
                "": self.carrier_risk_category,
                "": self.number_blocked,
                "": self.number_blocked_date,
                "": self.number_blocked_last_3_months,
                "": self.sms_pumping_risk_score,
                "": self.error_code,
            }

    class SmsPumpingRiskParameters(object):
        """
        :ivar partner_sub_id:
        """

        def __init__(self, payload: Dict[str, Any]):

            self.partner_sub_id: Optional[str] = payload.get("partner_sub_id")

        def to_dict(self):
            return {
                "partner_sub_id": self.partner_sub_id,
            }

    def __init__(self, version: Version):
        """
        Initialize the QueryList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/batch/query"

    def create(
        self, lookup_request1: Union[LookupRequest1, object] = values.unset
    ) -> QueryInstance:
        """
        Create the QueryInstance

        :param lookup_request1:

        :returns: The created QueryInstance
        """
        data = lookup_request1.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return QueryInstance(self._version, payload)

    async def create_async(
        self, lookup_request1: Union[LookupRequest1, object] = values.unset
    ) -> QueryInstance:
        """
        Asynchronously create the QueryInstance

        :param lookup_request1:

        :returns: The created QueryInstance
        """
        data = lookup_request1.to_dict()

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/json"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return QueryInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Lookups.V2.QueryList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Flex
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class PluginReleaseInstance(InstanceResource):
    """
    :ivar sid: The unique string that we created to identify the Plugin Release resource.
    :ivar account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Plugin Release resource and owns this resource.
    :ivar configuration_sid: The SID of the Plugin Configuration resource to release.
    :ivar date_created: The date and time in GMT when the Flex Plugin Release was created specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
    :ivar url: The absolute URL of the Plugin Release resource.
    """

    def __init__(
        self, version: Version, payload: Dict[str, Any], sid: Optional[str] = None
    ):
        super().__init__(version)

        self.sid: Optional[str] = payload.get("sid")
        self.account_sid: Optional[str] = payload.get("account_sid")
        self.configuration_sid: Optional[str] = payload.get("configuration_sid")
        self.date_created: Optional[datetime] = deserialize.iso8601_datetime(
            payload.get("date_created")
        )
        self.url: Optional[str] = payload.get("url")

        self._solution = {
            "sid": sid or self.sid,
        }
        self._context: Optional[PluginReleaseContext] = None

    @property
    def _proxy(self) -> "PluginReleaseContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: PluginReleaseContext for this PluginReleaseInstance
        """
        if self._context is None:
            self._context = PluginReleaseContext(
                self._version,
                sid=self._solution["sid"],
            )
        return self._context

    def fetch(
        self, flex_metadata: Union[str, object] = values.unset
    ) -> "PluginReleaseInstance":
        """
        Fetch the PluginReleaseInstance

        :param flex_metadata: The Flex-Metadata HTTP request header

        :returns: The fetched PluginReleaseInstance
        """
        return self._proxy.fetch(
            flex_metadata=flex_metadata,
        )

    async def fetch_async(
        self, flex_metadata: Union[str, object] = values.unset
    ) -> "PluginReleaseInstance":
        """
        Asynchronous coroutine to fetch the PluginReleaseInstance

        :param flex_metadata: The Flex-Metadata HTTP request header

        :returns: The fetched PluginReleaseInstance
        """
        return await self._proxy.fetch_async(
            flex_metadata=flex_metadata,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.FlexApi.V1.PluginReleaseInstance {}>".format(context)


class PluginReleaseContext(InstanceContext):

    def __init__(self, version: Version, sid: str):
        """
        Initialize the PluginReleaseContext

        :param version: Version that contains the resource
        :param sid: The SID of the Flex Plugin Release resource to fetch.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "sid": sid,
        }
        self._uri = "/PluginService/Releases/{sid}".format(**self._solution)

    def fetch(
        self, flex_metadata: Union[str, object] = values.unset
    ) -> PluginReleaseInstance:
        """
        Fetch the PluginReleaseInstance

        :param flex_metadata: The Flex-Metadata HTTP request header

        :returns: The fetched PluginReleaseInstance
        """

        data = values.of(
            {
                "Flex-Metadata": flex_metadata,
            }
        )

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(
            method="GET", uri=self._uri, params=data, headers=headers
        )

        return PluginReleaseInstance(
            self._version,
            payload,
            sid=self._solution["sid"],
        )

    async def fetch_async(
        self, flex_metadata: Union[str, object] = values.unset
    ) -> PluginReleaseInstance:
        """
        Asynchronous coroutine to fetch the PluginReleaseInstance

        :param flex_metadata: The Flex-Metadata HTTP request header

        :returns: The fetched PluginReleaseInstance
        """

        data = values.of(
            {
                "Flex-Metadata": flex_metadata,
            }
        )

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )

        return PluginReleaseInstance(
            self._version,
            payload,
            sid=self._solution["sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.FlexApi.V1.PluginReleaseContext {}>".format(context)


class PluginReleasePage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> PluginReleaseInstance:
        """
        Build an instance of PluginReleaseInstance

        :param payload: Payload response from the API
        """
        return PluginReleaseInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.FlexApi.V1.PluginReleasePage>"


class PluginReleaseList(ListResource):

    def __init__(self, version: Version):
        """
        Initialize the PluginReleaseList

        :param version: Version that contains the resource

        """
        super().__init__(version)

        self._uri = "/PluginService/Releases"

    def create(
        self, configuration_id: str, flex_metadata: Union[str, object] = values.unset
    ) -> PluginReleaseInstance:
        """
        Create the PluginReleaseInstance

        :param configuration_id: The SID or the Version of the Flex Plugin Configuration to release.
        :param flex_metadata: The Flex-Metadata HTTP request header

        :returns: The created PluginReleaseInstance
        """

        data = values.of(
            {
                "ConfigurationId": configuration_id,
            }
        )
        headers = values.of(
            {
                "Flex-Metadata": flex_metadata,
                "Content-Type": "application/x-www-form-urlencoded",
            }
        )

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return PluginReleaseInstance(self._version, payload)

    async def create_async(
        self, configuration_id: str, flex_metadata: Union[str, object] = values.unset
    ) -> PluginReleaseInstance:
        """
        Asynchronously create the PluginReleaseInstance

        :param configuration_id: The SID or the Version of the Flex Plugin Configuration to release.
        :param flex_metadata: The Flex-Metadata HTTP request header

        :returns: The created PluginReleaseInstance
        """

        data = values.of(
            {
                "ConfigurationId": configuration_id,
            }
        )
        headers = values.of(
            {
                "Flex-Metadata": flex_metadata,
                "Content-Type": "application/x-www-form-urlencoded",
            }
        )

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return PluginReleaseInstance(self._version, payload)

    def stream(
        self,
        flex_metadata: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[PluginReleaseInstance]:
        """
        Streams PluginReleaseInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str flex_metadata: The Flex-Metadata HTTP request header
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(flex_metadata=flex_metadata, page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        flex_metadata: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[PluginReleaseInstance]:
        """
        Asynchronously streams PluginReleaseInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str flex_metadata: The Flex-Metadata HTTP request header
        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            flex_metadata=flex_metadata, page_size=limits["page_size"]
        )

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        flex_metadata: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[PluginReleaseInstance]:
        """
        Lists PluginReleaseInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str flex_metadata: The Flex-Metadata HTTP request header
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                flex_metadata=flex_metadata,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        flex_metadata: Union[str, object] = values.unset,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[PluginReleaseInstance]:
        """
        Asynchronously lists PluginReleaseInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str flex_metadata: The Flex-Metadata HTTP request header
        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                flex_metadata=flex_metadata,
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        flex_metadata: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> PluginReleasePage:
        """
        Retrieve a single page of PluginReleaseInstance records from the API.
        Request is executed immediately

        :param flex_metadata: The Flex-Metadata HTTP request header
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of PluginReleaseInstance
        """
        data = values.of(
            {
                "Flex-Metadata": flex_metadata,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of(
            {
                "Flex-Metadata": flex_metadata,
                "Content-Type": "application/x-www-form-urlencoded",
            }
        )

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return PluginReleasePage(self._version, response)

    async def page_async(
        self,
        flex_metadata: Union[str, object] = values.unset,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> PluginReleasePage:
        """
        Asynchronously retrieve a single page of PluginReleaseInstance records from the API.
        Request is executed immediately

        :param flex_metadata: The Flex-Metadata HTTP request header
        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of PluginReleaseInstance
        """
        data = values.of(
            {
                "Flex-Metadata": flex_metadata,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of(
            {
                "Flex-Metadata": flex_metadata,
                "Content-Type": "application/x-www-form-urlencoded",
            }
        )

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return PluginReleasePage(self._version, response)

    def get_page(self, target_url: str) -> PluginReleasePage:
        """
        Retrieve a specific page of PluginReleaseInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of PluginReleaseInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return PluginReleasePage(self._version, response)

    async def get_page_async(self, target_url: str) -> PluginReleasePage:
        """
        Asynchronously retrieve a specific page of PluginReleaseInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of PluginReleaseInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return PluginReleasePage(self._version, response)

    def get(self, sid: str) -> PluginReleaseContext:
        """
        Constructs a PluginReleaseContext

        :param sid: The SID of the Flex Plugin Release resource to fetch.
        """
        return PluginReleaseContext(self._version, sid=sid)

    def __call__(self, sid: str) -> PluginReleaseContext:
        """
        Constructs a PluginReleaseContext

        :param sid: The SID of the Flex Plugin Release resource to fetch.
        """
        return PluginReleaseContext(self._version, sid=sid)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.FlexApi.V1.PluginReleaseList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Accounts
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Optional
from twilio.base.version import Version
from twilio.base.domain import Domain
from twilio.rest.accounts.v1.auth_token_promotion import AuthTokenPromotionList
from twilio.rest.accounts.v1.bulk_consents import BulkConsentsList
from twilio.rest.accounts.v1.bulk_contacts import BulkContactsList
from twilio.rest.accounts.v1.credential import CredentialList
from twilio.rest.accounts.v1.safelist import SafelistList
from twilio.rest.accounts.v1.secondary_auth_token import SecondaryAuthTokenList


class V1(Version):

    def __init__(self, domain: Domain):
        """
        Initialize the V1 version of Accounts

        :param domain: The Twilio.accounts domain
        """
        super().__init__(domain, "v1")
        self._auth_token_promotion: Optional[AuthTokenPromotionList] = None
        self._bulk_consents: Optional[BulkConsentsList] = None
        self._bulk_contacts: Optional[BulkContactsList] = None
        self._credentials: Optional[CredentialList] = None
        self._safelist: Optional[SafelistList] = None
        self._secondary_auth_token: Optional[SecondaryAuthTokenList] = None

    @property
    def auth_token_promotion(self) -> AuthTokenPromotionList:
        if self._auth_token_promotion is None:
            self._auth_token_promotion = AuthTokenPromotionList(self)
        return self._auth_token_promotion

    @property
    def bulk_consents(self) -> BulkConsentsList:
        if self._bulk_consents is None:
            self._bulk_consents = BulkConsentsList(self)
        return self._bulk_consents

    @property
    def bulk_contacts(self) -> BulkContactsList:
        if self._bulk_contacts is None:
            self._bulk_contacts = BulkContactsList(self)
        return self._bulk_contacts

    @property
    def credentials(self) -> CredentialList:
        if self._credentials is None:
            self._credentials = CredentialList(self)
        return self._credentials

    @property
    def safelist(self) -> SafelistList:
        if self._safelist is None:
            self._safelist = SafelistList(self)
        return self._safelist

    @property
    def secondary_auth_token(self) -> SecondaryAuthTokenList:
        if self._secondary_auth_token is None:
            self._secondary_auth_token = SecondaryAuthTokenList(self)
        return self._secondary_auth_token

    def __repr__(self) -> str:
        """
        Provide a friendly representation
        :returns: Machine friendly representation
        """
        return "<Twilio.Accounts.V1>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Api
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class MemberInstance(InstanceResource):
    """
    :ivar call_sid: The SID of the [Call](https://www.twilio.com/docs/voice/api/call-resource) the Member resource is associated with.
    :ivar date_enqueued: The date that the member was enqueued, given in RFC 2822 format.
    :ivar position: This member's current position in the queue.
    :ivar uri: The URI of the resource, relative to `https://api.twilio.com`.
    :ivar wait_time: The number of seconds the member has been in the queue.
    :ivar queue_sid: The SID of the Queue the member is in.
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        account_sid: str,
        queue_sid: str,
        call_sid: Optional[str] = None,
    ):
        super().__init__(version)

        self.call_sid: Optional[str] = payload.get("call_sid")
        self.date_enqueued: Optional[datetime] = deserialize.rfc2822_datetime(
            payload.get("date_enqueued")
        )
        self.position: Optional[int] = deserialize.integer(payload.get("position"))
        self.uri: Optional[str] = payload.get("uri")
        self.wait_time: Optional[int] = deserialize.integer(payload.get("wait_time"))
        self.queue_sid: Optional[str] = payload.get("queue_sid")

        self._solution = {
            "account_sid": account_sid,
            "queue_sid": queue_sid,
            "call_sid": call_sid or self.call_sid,
        }
        self._context: Optional[MemberContext] = None

    @property
    def _proxy(self) -> "MemberContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: MemberContext for this MemberInstance
        """
        if self._context is None:
            self._context = MemberContext(
                self._version,
                account_sid=self._solution["account_sid"],
                queue_sid=self._solution["queue_sid"],
                call_sid=self._solution["call_sid"],
            )
        return self._context

    def fetch(self) -> "MemberInstance":
        """
        Fetch the MemberInstance


        :returns: The fetched MemberInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self) -> "MemberInstance":
        """
        Asynchronous coroutine to fetch the MemberInstance


        :returns: The fetched MemberInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self, url: str, method: Union[str, object] = values.unset
    ) -> "MemberInstance":
        """
        Update the MemberInstance

        :param url: The absolute URL of the Queue resource.
        :param method: How to pass the update request data. Can be `GET` or `POST` and the default is `POST`. `POST` sends the data as encoded form data and `GET` sends the data as query parameters.

        :returns: The updated MemberInstance
        """
        return self._proxy.update(
            url=url,
            method=method,
        )

    async def update_async(
        self, url: str, method: Union[str, object] = values.unset
    ) -> "MemberInstance":
        """
        Asynchronous coroutine to update the MemberInstance

        :param url: The absolute URL of the Queue resource.
        :param method: How to pass the update request data. Can be `GET` or `POST` and the default is `POST`. `POST` sends the data as encoded form data and `GET` sends the data as query parameters.

        :returns: The updated MemberInstance
        """
        return await self._proxy.update_async(
            url=url,
            method=method,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.MemberInstance {}>".format(context)


class MemberContext(InstanceContext):

    def __init__(
        self, version: Version, account_sid: str, queue_sid: str, call_sid: str
    ):
        """
        Initialize the MemberContext

        :param version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Member resource(s) to update.
        :param queue_sid: The SID of the Queue in which to find the members to update.
        :param call_sid: The [Call](https://www.twilio.com/docs/voice/api/call-resource) SID of the resource(s) to update.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
            "queue_sid": queue_sid,
            "call_sid": call_sid,
        }
        self._uri = (
            "/Accounts/{account_sid}/Queues/{queue_sid}/Members/{call_sid}.json".format(
                **self._solution
            )
        )

    def fetch(self) -> MemberInstance:
        """
        Fetch the MemberInstance


        :returns: The fetched MemberInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = self._version.fetch(method="GET", uri=self._uri, headers=headers)

        return MemberInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
            call_sid=self._solution["call_sid"],
        )

    async def fetch_async(self) -> MemberInstance:
        """
        Asynchronous coroutine to fetch the MemberInstance


        :returns: The fetched MemberInstance
        """

        headers = values.of({})

        headers["Accept"] = "application/json"

        payload = await self._version.fetch_async(
            method="GET", uri=self._uri, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
            call_sid=self._solution["call_sid"],
        )

    def update(
        self, url: str, method: Union[str, object] = values.unset
    ) -> MemberInstance:
        """
        Update the MemberInstance

        :param url: The absolute URL of the Queue resource.
        :param method: How to pass the update request data. Can be `GET` or `POST` and the default is `POST`. `POST` sends the data as encoded form data and `GET` sends the data as query parameters.

        :returns: The updated MemberInstance
        """

        data = values.of(
            {
                "Url": url,
                "Method": method,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
            call_sid=self._solution["call_sid"],
        )

    async def update_async(
        self, url: str, method: Union[str, object] = values.unset
    ) -> MemberInstance:
        """
        Asynchronous coroutine to update the MemberInstance

        :param url: The absolute URL of the Queue resource.
        :param method: How to pass the update request data. Can be `GET` or `POST` and the default is `POST`. `POST` sends the data as encoded form data and `GET` sends the data as query parameters.

        :returns: The updated MemberInstance
        """

        data = values.of(
            {
                "Url": url,
                "Method": method,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return MemberInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
            call_sid=self._solution["call_sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.MemberContext {}>".format(context)


class MemberPage(Page):

    def get_instance(self, payload: Dict[str, Any]) -> MemberInstance:
        """
        Build an instance of MemberInstance

        :param payload: Payload response from the API
        """
        return MemberInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.MemberPage>"


class MemberList(ListResource):

    def __init__(self, version: Version, account_sid: str, queue_sid: str):
        """
        Initialize the MemberList

        :param version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Member resource(s) to read.
        :param queue_sid: The SID of the Queue in which to find the members

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
            "queue_sid": queue_sid,
        }
        self._uri = "/Accounts/{account_sid}/Queues/{queue_sid}/Members.json".format(
            **self._solution
        )

    def stream(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[MemberInstance]:
        """
        Streams MemberInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[MemberInstance]:
        """
        Asynchronously streams MemberInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(page_size=limits["page_size"])

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[MemberInstance]:
        """
        Lists MemberInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[MemberInstance]:
        """
        Asynchronously lists MemberInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> MemberPage:
        """
        Retrieve a single page of MemberInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of MemberInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return MemberPage(self._version, response, self._solution)

    async def page_async(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> MemberPage:
        """
        Asynchronously retrieve a single page of MemberInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of MemberInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return MemberPage(self._version, response, self._solution)

    def get_page(self, target_url: str) -> MemberPage:
        """
        Retrieve a specific page of MemberInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of MemberInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return MemberPage(self._version, response, self._solution)

    async def get_page_async(self, target_url: str) -> MemberPage:
        """
        Asynchronously retrieve a specific page of MemberInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of MemberInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return MemberPage(self._version, response, self._solution)

    def get(self, call_sid: str) -> MemberContext:
        """
        Constructs a MemberContext

        :param call_sid: The [Call](https://www.twilio.com/docs/voice/api/call-resource) SID of the resource(s) to update.
        """
        return MemberContext(
            self._version,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
            call_sid=call_sid,
        )

    def __call__(self, call_sid: str) -> MemberContext:
        """
        Constructs a MemberContext

        :param call_sid: The [Call](https://www.twilio.com/docs/voice/api/call-resource) SID of the resource(s) to update.
        """
        return MemberContext(
            self._version,
            account_sid=self._solution["account_sid"],
            queue_sid=self._solution["queue_sid"],
            call_sid=call_sid,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.MemberList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Flex
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""

from typing import Any, Dict, List, Optional, Union, Iterator, AsyncIterator
from twilio.base import serialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class InteractionChannelParticipantInstance(InstanceResource):

    class Status(object):
        CLOSED = "closed"
        WRAPUP = "wrapup"

    class Type(object):
        SUPERVISOR = "supervisor"
        CUSTOMER = "customer"
        EXTERNAL = "external"
        AGENT = "agent"
        UNKNOWN = "unknown"

    """
    :ivar sid: The unique string created by Twilio to identify an Interaction Channel Participant resource.
    :ivar type: 
    :ivar interaction_sid: The Interaction Sid for this channel.
    :ivar channel_sid: The Channel Sid for this Participant.
    :ivar url: 
    :ivar routing_properties: The Participant's routing properties.
    """

    def __init__(
        self,
        version: Version,
        payload: Dict[str, Any],
        interaction_sid: str,
        channel_sid: str,
        sid: Optional[str] = None,
    ):
        super().__init__(version)

        self.sid: Optional[str] = payload.get("sid")
        self.type: Optional["InteractionChannelParticipantInstance.Type"] = payload.get(
            "type"
        )
        self.interaction_sid: Optional[str] = payload.get("interaction_sid")
        self.channel_sid: Optional[str] = payload.get("channel_sid")
        self.url: Optional[str] = payload.get("url")
        self.routing_properties: Optional[Dict[str, object]] = payload.get(
            "routing_properties"
        )

        self._solution = {
            "interaction_sid": interaction_sid,
            "channel_sid": channel_sid,
            "sid": sid or self.sid,
        }
        self._context: Optional[InteractionChannelParticipantContext] = None

    @property
    def _proxy(self) -> "InteractionChannelParticipantContext":
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: InteractionChannelParticipantContext for this InteractionChannelParticipantInstance
        """
        if self._context is None:
            self._context = InteractionChannelParticipantContext(
                self._version,
                interaction_sid=self._solution["interaction_sid"],
                channel_sid=self._solution["channel_sid"],
                sid=self._solution["sid"],
            )
        return self._context

    def update(
        self, status: "InteractionChannelParticipantInstance.Status"
    ) -> "InteractionChannelParticipantInstance":
        """
        Update the InteractionChannelParticipantInstance

        :param status:

        :returns: The updated InteractionChannelParticipantInstance
        """
        return self._proxy.update(
            status=status,
        )

    async def update_async(
        self, status: "InteractionChannelParticipantInstance.Status"
    ) -> "InteractionChannelParticipantInstance":
        """
        Asynchronous coroutine to update the InteractionChannelParticipantInstance

        :param status:

        :returns: The updated InteractionChannelParticipantInstance
        """
        return await self._proxy.update_async(
            status=status,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.FlexApi.V1.InteractionChannelParticipantInstance {}>".format(
            context
        )


class InteractionChannelParticipantContext(InstanceContext):

    def __init__(
        self, version: Version, interaction_sid: str, channel_sid: str, sid: str
    ):
        """
        Initialize the InteractionChannelParticipantContext

        :param version: Version that contains the resource
        :param interaction_sid: The Interaction Sid for this channel.
        :param channel_sid: The Channel Sid for this Participant.
        :param sid: The unique string created by Twilio to identify an Interaction Channel resource.
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "interaction_sid": interaction_sid,
            "channel_sid": channel_sid,
            "sid": sid,
        }
        self._uri = "/Interactions/{interaction_sid}/Channels/{channel_sid}/Participants/{sid}".format(
            **self._solution
        )

    def update(
        self, status: "InteractionChannelParticipantInstance.Status"
    ) -> InteractionChannelParticipantInstance:
        """
        Update the InteractionChannelParticipantInstance

        :param status:

        :returns: The updated InteractionChannelParticipantInstance
        """

        data = values.of(
            {
                "Status": status,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.update(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return InteractionChannelParticipantInstance(
            self._version,
            payload,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=self._solution["sid"],
        )

    async def update_async(
        self, status: "InteractionChannelParticipantInstance.Status"
    ) -> InteractionChannelParticipantInstance:
        """
        Asynchronous coroutine to update the InteractionChannelParticipantInstance

        :param status:

        :returns: The updated InteractionChannelParticipantInstance
        """

        data = values.of(
            {
                "Status": status,
            }
        )
        headers = values.of({})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.update_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return InteractionChannelParticipantInstance(
            self._version,
            payload,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=self._solution["sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.FlexApi.V1.InteractionChannelParticipantContext {}>".format(
            context
        )


class InteractionChannelParticipantPage(Page):

    def get_instance(
        self, payload: Dict[str, Any]
    ) -> InteractionChannelParticipantInstance:
        """
        Build an instance of InteractionChannelParticipantInstance

        :param payload: Payload response from the API
        """
        return InteractionChannelParticipantInstance(
            self._version,
            payload,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.FlexApi.V1.InteractionChannelParticipantPage>"


class InteractionChannelParticipantList(ListResource):

    def __init__(self, version: Version, interaction_sid: str, channel_sid: str):
        """
        Initialize the InteractionChannelParticipantList

        :param version: Version that contains the resource
        :param interaction_sid: The Interaction Sid for this channel.
        :param channel_sid: The Channel Sid for this Participant.

        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "interaction_sid": interaction_sid,
            "channel_sid": channel_sid,
        }
        self._uri = "/Interactions/{interaction_sid}/Channels/{channel_sid}/Participants".format(
            **self._solution
        )

    def create(
        self,
        type: "InteractionChannelParticipantInstance.Type",
        media_properties: object,
        routing_properties: Union[object, object] = values.unset,
    ) -> InteractionChannelParticipantInstance:
        """
        Create the InteractionChannelParticipantInstance

        :param type:
        :param media_properties: JSON representing the Media Properties for the new Participant.
        :param routing_properties: Object representing the Routing Properties for the new Participant.

        :returns: The created InteractionChannelParticipantInstance
        """

        data = values.of(
            {
                "Type": type,
                "MediaProperties": serialize.object(media_properties),
                "RoutingProperties": serialize.object(routing_properties),
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = self._version.create(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return InteractionChannelParticipantInstance(
            self._version,
            payload,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
        )

    async def create_async(
        self,
        type: "InteractionChannelParticipantInstance.Type",
        media_properties: object,
        routing_properties: Union[object, object] = values.unset,
    ) -> InteractionChannelParticipantInstance:
        """
        Asynchronously create the InteractionChannelParticipantInstance

        :param type:
        :param media_properties: JSON representing the Media Properties for the new Participant.
        :param routing_properties: Object representing the Routing Properties for the new Participant.

        :returns: The created InteractionChannelParticipantInstance
        """

        data = values.of(
            {
                "Type": type,
                "MediaProperties": serialize.object(media_properties),
                "RoutingProperties": serialize.object(routing_properties),
            }
        )
        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Content-Type"] = "application/x-www-form-urlencoded"

        headers["Accept"] = "application/json"

        payload = await self._version.create_async(
            method="POST", uri=self._uri, data=data, headers=headers
        )

        return InteractionChannelParticipantInstance(
            self._version,
            payload,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
        )

    def stream(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> Iterator[InteractionChannelParticipantInstance]:
        """
        Streams InteractionChannelParticipantInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> AsyncIterator[InteractionChannelParticipantInstance]:
        """
        Asynchronously streams InteractionChannelParticipantInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param limit: Upper limit for the number of records to return. stream()
                      guarantees to never return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, stream() will attempt to read the
                          limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(page_size=limits["page_size"])

        return self._version.stream_async(page, limits["limit"])

    def list(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[InteractionChannelParticipantInstance]:
        """
        Lists InteractionChannelParticipantInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return list(
            self.stream(
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[InteractionChannelParticipantInstance]:
        """
        Asynchronously lists InteractionChannelParticipantInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param limit: Upper limit for the number of records to return. list() guarantees
                      never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
                          the default value of 50 records.  If no page_size is defined
                          but a limit is defined, list() will attempt to read the limit
                          with the most efficient page size, i.e. min(limit, 1000)

        :returns: list that will contain up to limit results
        """
        return [
            record
            async for record in await self.stream_async(
                limit=limit,
                page_size=page_size,
            )
        ]

    def page(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> InteractionChannelParticipantPage:
        """
        Retrieve a single page of InteractionChannelParticipantInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of InteractionChannelParticipantInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = self._version.page(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return InteractionChannelParticipantPage(
            self._version, response, self._solution
        )

    async def page_async(
        self,
        page_token: Union[str, object] = values.unset,
        page_number: Union[int, object] = values.unset,
        page_size: Union[int, object] = values.unset,
    ) -> InteractionChannelParticipantPage:
        """
        Asynchronously retrieve a single page of InteractionChannelParticipantInstance records from the API.
        Request is executed immediately

        :param page_token: PageToken provided by the API
        :param page_number: Page Number, this value is simply for client state
        :param page_size: Number of records to return, defaults to 50

        :returns: Page of InteractionChannelParticipantInstance
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        headers = values.of({"Content-Type": "application/x-www-form-urlencoded"})

        headers["Accept"] = "application/json"

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data, headers=headers
        )
        return InteractionChannelParticipantPage(
            self._version, response, self._solution
        )

    def get_page(self, target_url: str) -> InteractionChannelParticipantPage:
        """
        Retrieve a specific page of InteractionChannelParticipantInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of InteractionChannelParticipantInstance
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return InteractionChannelParticipantPage(
            self._version, response, self._solution
        )

    async def get_page_async(
        self, target_url: str
    ) -> InteractionChannelParticipantPage:
        """
        Asynchronously retrieve a specific page of InteractionChannelParticipantInstance records from the API.
        Request is executed immediately

        :param target_url: API-generated URL for the requested results page

        :returns: Page of InteractionChannelParticipantInstance
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return InteractionChannelParticipantPage(
            self._version, response, self._solution
        )

    def get(self, sid: str) -> InteractionChannelParticipantContext:
        """
        Constructs a InteractionChannelParticipantContext

        :param sid: The unique string created by Twilio to identify an Interaction Channel resource.
        """
        return InteractionChannelParticipantContext(
            self._version,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=sid,
        )

    def __call__(self, sid: str) -> InteractionChannelParticipantContext:
        """
        Constructs a InteractionChannelParticipantContext

        :param sid: The unique string created by Twilio to identify an Interaction Channel resource.
        """
        return InteractionChannelParticipantContext(
            self._version,
            interaction_sid=self._solution["interaction_sid"],
            channel_sid=self._solution["channel_sid"],
            sid=sid,
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.FlexApi.V1.InteractionChannelParticipantList>"
